#include <stdint.h>
#include <stdbool.h>

#include "Controller.h"
#include "Menu.h"
#include "Printf.h"
#include "Buttons.h"
#include "ADC.h"
#include "Configuration.h"
#include "PWMController.h"
#include "NTC.h"

static struct controllerState internalState;

unsigned char controllerGetMode(void) {
	return internalState.mode;
}

unsigned char controllerGetCurPercent(void) {
	return internalState.curPercent;
}

unsigned char controllerGetMaxPercent(void) {
	return internalState.configuration.maxMotion;
}

unsigned char controllerGetMotorTemperature(void) {
	return internalState.temperature1;
}

unsigned char controllerGetMosfetTemperature(void) {
	return internalState.temperature2;
}

uint16_t controllerGetCalibrationValue(void) {
	return internalState.calValue;
}

uint16_t controllerGetCalibrationLow(void) {
	return internalState.calMinValue;
}

uint16_t controllerGetCalibrationHigh(void) {
	return internalState.calMaxValue;
}

unsigned char controllerGetSource(void) {
	return internalState.configuration.pwmSource;
}

void initController(void) {
	internalState.screen = SCREEN_MAIN;
	internalState.mode = MODE_STOPPED;
	internalState.curPercent = 0;
	internalState.forceRedraw = true;
	configurationGet(&internalState.configuration);
}

static void inputGetValue(struct inputResults *aInputResults, unsigned char aSampleSource, uint16_t aMinValue, uint16_t aMaxValue) {
	uint16_t currentValue = adcSample(aSampleSource);
	aInputResults->range = aMaxValue - aMinValue;
	if (currentValue < aMinValue) {
		currentValue = aMinValue;
	} else if (currentValue > aMaxValue) {
		currentValue = aMaxValue;
	}
	currentValue -= aMinValue;
	aInputResults->value = currentValue;
}

static void inputGetDeviceValue(struct inputResults *aInputResults, unsigned char aSampleSource) {
	if (aSampleSource == ADCINPUT_POTENTIOMETER) {
		inputGetValue(aInputResults, aSampleSource, internalState.configuration.potMinValue, internalState.configuration.potMaxValue);
	} else {
		inputGetValue(aInputResults, aSampleSource, internalState.configuration.pedalMinValue, internalState.configuration.pedalMaxValue);
	}
}

static uint16_t scaleInput(const struct inputResults *aInputResults, uint16_t aMaxValue) {
	return ((uint32_t)aInputResults->value) * (uint32_t)aMaxValue / ((uint32_t)aInputResults->range);
}

void controllerLoop(void) {
	if (internalState.screen == SCREEN_MAIN) {
		struct inputResults input;
		if (buttonGet(0)) {	
			if (internalState.mode == MODE_STOPPED) {
				internalState.mode = MODE_STARTED;				
			} else {
				internalState.mode = MODE_STOPPED;
				pwmSetRawValue(0);
			}
		}
		internalState.temperature1 = adcValueToTemperature(adcSample(ADCINPUT_NTC_MOTOR));
		internalState.temperature2 = adcValueToTemperature(adcSample(ADCINPUT_NTC_MOSFET));
		inputGetDeviceValue(&input, (controllerGetSource() == PWMSOURCE_PEDAL) ? ADCINPUT_PEDAL : ADCINPUT_POTENTIOMETER);
		internalState.curPercent = scaleInput(&input, controllerGetMaxPercent());
		if (internalState.mode == MODE_STARTED) {
			pwmSetRawValue(scaleInput(&input, pwmGetMaxValue() * controllerGetMaxPercent() / 100));
		}
		updateScreenMain(internalState.forceRedraw);
		internalState.forceRedraw = false;
	
		if (buttonGet(1)) {
			internalState.screen = SCREEN_SELECT_CAL_PEDAL;
			internalState.forceRedraw = true;
			pwmSetRawValue(0);
		}
	} else if (internalState.screen == SCREEN_SELECT_CAL_PEDAL) {
		updateScreenSelectCalPedal(internalState.forceRedraw);
		internalState.forceRedraw = false;

		if (buttonGet(0)) {
			internalState.calMinValue = 65535;
			internalState.calMaxValue = 0;
			internalState.screen = SCREEN_CAL_PEDAL;
			internalState.forceRedraw = true;
		} else if (buttonGet(1)) {
			internalState.screen = SCREEN_SELECT_CAL_POT;
			internalState.forceRedraw = true;
		}
	} else if (internalState.screen == SCREEN_SELECT_CAL_POT) {
		updateScreenSelectCalPot(internalState.forceRedraw);
		internalState.forceRedraw = false;

		if (buttonGet(0)) {
			internalState.calMinValue = 65535;
			internalState.calMaxValue = 0;
			internalState.screen = SCREEN_CAL_POT;
			internalState.forceRedraw = true;
		} else if (buttonGet(1)) {
			internalState.screen = SCREEN_SELECT_MAX;
			internalState.forceRedraw = true;
		}
	} else if (internalState.screen == SCREEN_SELECT_MAX) {
		updateScreenSelectMax(internalState.forceRedraw);
		internalState.forceRedraw = false;

		if (buttonGet(0)) {
			internalState.screen = SCREEN_MAX;
			internalState.forceRedraw = true;
		} else if (buttonGet(1)) {
			internalState.screen = SCREEN_SELECT_SOURCE;
			internalState.forceRedraw = true;
		}
	} else if (internalState.screen == SCREEN_SELECT_SOURCE) {
		updateScreenSelectSource(internalState.forceRedraw);
		internalState.forceRedraw = false;

		if (buttonGet(0)) {
			if (internalState.configuration.pwmSource == PWMSOURCE_PEDAL) {
				internalState.configuration.pwmSource = PWMSOURCE_MOTOR;
			} else {
				internalState.configuration.pwmSource = PWMSOURCE_PEDAL;
			}			
			configurationSet(&internalState.configuration);
		} else if (buttonGet(1)) {
			internalState.screen = SCREEN_INFO;
			internalState.forceRedraw = true;
		}
	} else if (internalState.screen == SCREEN_INFO) {
		updateScreenInfos(internalState.forceRedraw);
		internalState.forceRedraw = false;

		if (buttonGet(1)) {
			internalState.screen = SCREEN_MAIN;
			internalState.forceRedraw = true;
		}
	} else if (internalState.screen == SCREEN_CAL_PEDAL) {
		uint16_t currentValue = adcSample(ADCINPUT_PEDAL);
		if (currentValue < internalState.calMinValue) {
			internalState.calMinValue = currentValue;
		}
		if (currentValue > internalState.calMaxValue) {
			internalState.calMaxValue = currentValue;
		}

		updateScreenCalPedal(internalState.forceRedraw);
		internalState.forceRedraw = false;
		
		if (buttonGet(0)) {
			internalState.configuration.pedalMinValue = internalState.calMinValue;
			internalState.configuration.pedalMaxValue = internalState.calMaxValue;
			configurationSet(&internalState.configuration);
			internalState.screen = SCREEN_SELECT_CAL_PEDAL;
			internalState.forceRedraw = true;
		}
	} else if (internalState.screen == SCREEN_CAL_POT) {
		uint16_t currentValue = adcSample(ADCINPUT_POTENTIOMETER);
		if (currentValue < internalState.calMinValue) {
			internalState.calMinValue = currentValue;
		}
		if (currentValue > internalState.calMaxValue) {
			internalState.calMaxValue = currentValue;
		}
		
		updateScreenCalPot(internalState.forceRedraw);
		internalState.forceRedraw = false;
		
		if (buttonGet(0)) {
			internalState.configuration.potMinValue = internalState.calMinValue;
			internalState.configuration.potMaxValue = internalState.calMaxValue;
			configurationSet(&internalState.configuration);
			internalState.screen = SCREEN_SELECT_CAL_POT;
			internalState.forceRedraw = true;
		}
	} else if (internalState.screen == SCREEN_MAX) {
		struct inputResults input;
		inputGetDeviceValue(&input, ADCINPUT_POTENTIOMETER);		
		internalState.calValue = scaleInput(&input, 100);
		
		updateScreenMaxLeft(internalState.forceRedraw);
		internalState.forceRedraw = false;
		
		if (buttonGet(0)) {
			internalState.configuration.maxMotion = internalState.calValue;
			configurationSet(&internalState.configuration);
			internalState.screen = SCREEN_SELECT_MAX;
			internalState.forceRedraw = true;
		}
	}
}

