#include <stdio.h>
#include <avr/io.h>
#include <util/delay.h>
#include <avr/pgmspace.h>

#include "HAL.h"
#include "LCD.h"
#include "Delay.h"

#ifdef LCD_PRINTF
static int lcdPutChar(char, FILE*);
FILE lcdout = FDEV_SETUP_STREAM(lcdPutChar, NULL, _FDEV_SETUP_WRITE);
#endif
static volatile struct LCDDisplayState lcdDisplayState;

/* Control-Register: RS = LOW */
/* Data-Register   : RS = HIGH */
static void lcdSetRegister(unsigned char aRegister) {
	LCD_RS_SetConditional(aRegister);
}

static void lcdSetDataBusModeRead(void) {
	/* Bring the data pins into reading position */
	LCD_D4_ModeInput();
	LCD_D5_ModeInput();
	LCD_D6_ModeInput();
	LCD_D7_ModeInput();
	
	/* Disable internal pullups */
	LCD_D4_SetPullupInactive();
	LCD_D5_SetPullupInactive();
	LCD_D6_SetPullupInactive();
	LCD_D7_SetPullupInactive();

	LCD_RW_SetActive();
}

static void lcdSetDataBusModeWrite(void) {
	LCD_RW_SetInactive();
	
	/* Bring the data pins into writing position */
	LCD_D4_ModeOutput();
	LCD_D5_ModeOutput();
	LCD_D6_ModeOutput();
	LCD_D7_ModeOutput();
}

static unsigned char lcdDataBusReadHalfword(unsigned char aDisplayNumber, unsigned char aRegister) {
	unsigned char returnValue;

	lcdSetDataBusModeRead();
	lcdSetRegister(aRegister);

#if LCD_DISPLAY_COUNT == 1
	LCD_E_SetActive();
#elif LCD_DISPLAY_COUNT == 2
	if (aDisplayNumber == 0) {
		LCD_E1_SetActive();
	} else {
		LCD_E2_SetActive();
	}
#else
	#error "Unsupported display count."
#endif

	delayMillis(1);			/* This delay is necessary, otherwise weird (!) things will happen */
	returnValue = (LCD_D7_GetBit() << 3) |
				(LCD_D6_GetBit() << 2) |
				(LCD_D5_GetBit() << 1) |
				(LCD_D4_GetBit() << 0);
	
#if LCD_DISPLAY_COUNT == 1
	LCD_E_SetInactive();
#elif LCD_DISPLAY_COUNT == 2
	if (aDisplayNumber == 0) {
		LCD_E1_SetInactive();
	} else {
		LCD_E2_SetInactive();
	}
#else
	#error "Unsupported display count."
#endif

	return returnValue;
}

static unsigned char lcdDataBusReadWord(unsigned char aDisplayNumber, unsigned char aRegister) {
	unsigned char word;
	word = lcdDataBusReadHalfword(aDisplayNumber, aRegister);
	word <<= 4;
	word |= lcdDataBusReadHalfword(aDisplayNumber, aRegister);
	return word;
}

static unsigned char lcdIsBusy(unsigned char aDisplayNumber) {
	return lcdDataBusReadWord(aDisplayNumber, REGISTER_CTRL) & 0x80;
}

static void lcdBusyWait(unsigned char aDisplayNumber) {
	if (lcdDisplayState.checkBusy) {
		while (lcdIsBusy(aDisplayNumber));
	}
}


static void lcdWriteHalfword(unsigned char aDisplayNumber, unsigned char aRegister, unsigned char aData) {
	lcdSetDataBusModeWrite();
	lcdSetRegister(aRegister);
	
	LCD_D4_SetConditional(aData & 0x01);
	LCD_D5_SetConditional(aData & 0x02);
	LCD_D6_SetConditional(aData & 0x04);
	LCD_D7_SetConditional(aData & 0x08);
	
#if LCD_DISPLAY_COUNT == 1
	LCD_E_SetActive();
	delayMillis(1);
	LCD_E_SetInactive();
#elif LCD_DISPLAY_COUNT == 2
	if (aDisplayNumber == 0) {
		LCD_E1_SetActive();
		delayMillis(1);
		LCD_E1_SetInactive();
	} else {
		LCD_E2_SetActive();
		delayMillis(1);
		LCD_E2_SetInactive();
	}
#else
	#error "Unsupported display count."
#endif

	delayMillis(1);
}

static void lcdWriteWord(unsigned char aDisplayNumber, unsigned char aRegister, unsigned char Data) {
	lcdBusyWait(aDisplayNumber);
	lcdWriteHalfword(aDisplayNumber, aRegister, Data >> 4);
	lcdBusyWait(aDisplayNumber);
	lcdWriteHalfword(aDisplayNumber, aRegister, Data);
}

/***************************************************************************/

void lcdClearDisplay(unsigned char aDisplayNumber) {
	lcdWriteWord(aDisplayNumber, REGISTER_CTRL, 0x01);
}

static void lcdCursorHome(unsigned char aDisplayNumber) {
	lcdWriteWord(aDisplayNumber, REGISTER_CTRL, 0x02);
}

static void lcdEntryModeSet(unsigned char aDisplayNumber, unsigned char IncrementDecrement, unsigned char Move) {
	lcdWriteWord(aDisplayNumber, REGISTER_CTRL, 0x04 | ((IncrementDecrement & 0x01) << 1) | ((Move & 0x01) << 0));
}

static void lcdDisplayOnOff(unsigned char aDisplayNumber, unsigned char DisplayOnOff, unsigned char Underline_Cursor, unsigned char Blinking_Cursor) {
	lcdWriteWord(aDisplayNumber, REGISTER_CTRL, 0x08 | ((DisplayOnOff & 0x01) << 2) | ((Underline_Cursor & 0x01) << 1) | ((Blinking_Cursor & 0x01) << 0));
}

static void lcdSystemSet(unsigned char aDisplayNumber, unsigned char BitMode, unsigned char NumberLines, unsigned char DrawingBox, unsigned char Inverter) {
	lcdWriteWord(aDisplayNumber, REGISTER_CTRL, 0x20 | ((BitMode & 0x01) << 4) | ((NumberLines & 0x01) << 3) | ((DrawingBox & 0x01) << 2) | ((Inverter & 0x01) << 1));
}

static void lcdSetDdRamAddress(unsigned char aDisplayNumber, unsigned char Address) {
	lcdWriteWord(aDisplayNumber, REGISTER_CTRL, 0x80 | Address);
}

static void resetSpecificLCD(unsigned char aDisplayNumber) {
	delayMillis(16);											/* 15ms */

	lcdDisplayState.checkBusy = 0;

	lcdWriteHalfword(aDisplayNumber, REGISTER_CTRL, 3);
	delayMillis(5);												/* 4,1ms */
	lcdWriteHalfword(aDisplayNumber, REGISTER_CTRL, 3);
	delayMillis(1);												/* 100µs */
	lcdWriteHalfword(aDisplayNumber, REGISTER_CTRL, 3);
	delayMillis(5);												/* 4,1ms */
	lcdWriteHalfword(aDisplayNumber, REGISTER_CTRL, 2);			/* System set: 4-bit-mode */
	
	lcdDisplayState.checkBusy = 1;
	lcdSystemSet(aDisplayNumber, SYSTEMSET_4BIT, SYSTEMSET_MULTILINE, SYSTEMSET_5_TO_7_DRAWINGBOX, SYSTEMSET_INVERTER_OFF);
	lcdDisplayOnOff(aDisplayNumber, DISPLAYONOFF_ON, DISPLAYONOFF_NOUNDERLINECURSOR, DISPLAYONOFF_NOBLINKINGCURSOR);
	lcdClearDisplay(aDisplayNumber);
	lcdCursorHome(aDisplayNumber);
	lcdEntryModeSet(aDisplayNumber, ENTRYMODESET_INCREMENT, ENTRYMODESET_NOMOVE);
}

void lcdInit(void) {
	unsigned char i;
	for (i = 0; i < LCD_DISPLAY_COUNT; i++) {
		resetSpecificLCD(i);
	}
}

void lcdGotoXY(unsigned char aX, unsigned char aY) {
	if (aY <= 1) {
		lcdDisplayState.lastDisplay = 0;
	} else {
		lcdDisplayState.lastDisplay = 1;
	}
	if ((aY == 1) || (aY == 3)) {
		aX += 0x40;
	}
	lcdSetDdRamAddress(lcdDisplayState.lastDisplay, aX);
}

void lcdPrintChar(char aCharacter) {
	lcdWriteWord(lcdDisplayState.lastDisplay, REGISTER_DATA, aCharacter);
}

#ifndef LCD_PRINTF
void lcdPrintCharXY(unsigned char aX, unsigned char aY, char aCharacter) {
	lcdGotoXY(aX, aY);
	lcdPrintChar(aCharacter);
}

void lcdPrintString_P(PGM_P aPgmPtr) {
	unsigned char c;
	while ((c = pgm_read_byte(aPgmPtr))) {
		lcdPrintChar(c);
		aPgmPtr++;
	}
}

void lcdPrintStringXY_P(unsigned char aX, unsigned char aY, PGM_P aPgmPtr) {
	lcdGotoXY(aX, aY);
	lcdPrintString_P(aPgmPtr);
}

#else

static int lcdPutChar(char aChar, FILE *aStream) {
	lcdPrintChar(aChar);
	return 1;
}

#endif
